/* BFD library support routines for the MSP architecture.
   Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
   Contributed by Dmitry Diky <diwil@mail.ru>

This file is part of BFD, the Binary File Descriptor library.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "bfd.h"
#include "sysdep.h"
#include "libbfd.h"

static const bfd_arch_info_type *compatible
  PARAMS ((const bfd_arch_info_type *, const bfd_arch_info_type *));

#define N(addr_bits, machine, print, default, next)		\
{								\
  16,				/* 16 bits in a word */		\
  addr_bits,			/* bits in an address */	\
  8,				/* 8 bits in a byte */		\
  bfd_arch_msp430,							\
  machine,			/* machine */			\
  "msp430",			/* arch_name  */		\
  print,			/* printable name */		\
  1,				/* section align power */	\
  default,			/* the default machine */	\
  compatible,							\
  bfd_default_scan,						\
  next								\
}

static const bfd_arch_info_type arch_info_struct[] =
{
  /* msp430x11x */
  N (16, bfd_mach_msp11, "msp:11", false, & arch_info_struct[1]),

  /* msp430x12x */
  N (16, bfd_mach_msp12, "msp:12", false, & arch_info_struct[2]),

  /* msp430x13x */
  N (16, bfd_mach_msp13, "msp:13", false, & arch_info_struct[3]),

  /* msp430x14x */
  N (16, bfd_mach_msp14, "msp:14", false, & arch_info_struct[4]),

  /* msp430x31x */
  N (16, bfd_mach_msp31, "msp:31", false, & arch_info_struct[5]), 

  /* msp430x32x */
  N (16, bfd_mach_msp32, "msp:32", false, & arch_info_struct[6]), 

  /* msp430x33x */
  N (16, bfd_mach_msp33, "msp:33", false, & arch_info_struct[7]),
  
  /* msp430x41x */
  N (16, bfd_mach_msp41, "msp:41", false, & arch_info_struct[8]),
  
  /* msp430x11x1 */
  N (16, bfd_mach_msp110, "msp:110", false, NULL)

};

const bfd_arch_info_type bfd_msp430_arch =
  N (16, bfd_mach_msp14, "msp:14", true, & arch_info_struct[0]);

/* This routine is provided two arch_infos and works out which AVR
   machine which would be compatible with both and returns a pointer
   to its info structure.  */

static const bfd_arch_info_type *
compatible (a,b)
     const bfd_arch_info_type * a;
     const bfd_arch_info_type * b;
{
  /* If a & b are for different architectures we can do nothing.  */
  if (a->arch != b->arch)
    return NULL;

  if (a->mach <= b->mach)
    return b;

  if (a->mach >= b->mach)
    return a;

  /* Never reached!  */
  return NULL;
}
