/* Parameters for execution on a msp430 series micro controllers.
    Copyright 2001 (c) Dmitry Diky diwil@mail.ru

 This file is part of GDB.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#ifdef __STDC__
 struct frame_info;
 struct frame_saved_regs;
 struct value;
 struct type;
#endif

typedef unsigned short INSN_WORD;


#define IEEE_FLOAT 1

 /* Define the bit, byte, and word ordering of the machine.  */
#define TARGET_BYTE_ORDER LITTLE_ENDIAN
#undef TARGET_INT_BIT
#define TARGET_INT_BIT	16
#undef TARGET_LONG_BIT
#define TARGET_LONG_BIT	32
#undef TARGET_PTR_BIT
#define TARGET_PTR_BIT	16

#define FUNCTION_START_OFFSET 0

/* A C expression that returns the address of the "real" code beyond
     the function entry prologue found at PC. */
extern CORE_ADDR msp430_skip_prologue PARAMS ((CORE_ADDR ip));
#define SKIP_PROLOGUE(ip) (msp430_skip_prologue(ip))

 /* Immediately after a function call, return the saved pc.
    Can't always go through the frames for this because on some machines
    the new frame is not set up until the new function executes
    some instructions.  */
CORE_ADDR msp430_saved_pc_after_call (struct frame_info * frame);
#define SAVED_PC_AFTER_CALL(frame) msp430_saved_pc_after_call (frame)

/* Stack grows downward.  */
#define INNER_THAN(lhs,rhs) ((lhs) < (rhs))

/* Do not execute insn at breakpoint */
#define DECR_PC_AFTER_BREAK 0

/* registers */
#define SREG_REGNUM 	2
#define SP_REGNUM 	1
#define PC_REGNUM	0


/* registers a word (16-bits) long */
#define REGISTER_SIZE	2
/* all regs are there, no phantoms */
#define NUM_REGS 	16

/* The total amount of space needed to store GDB's copy of the
     machine's register state */
#define REGISTER_BYTES	32

/* index within registers. */
#define REGISTER_BYTE(N)	(2*N)

#define REGISTER_RAW_SIZE(N)		2
#define REGISTER_VIRTUAL_SIZE(N)	2

#define MAX_REGISTER_RAW_SIZE		2
#define MAX_REGISTER_VIRTUAL_SIZE	8	/* can handle long long */

#define REGISTER_VIRTUAL_TYPE(N)	(builtin_type_unsigned_short)

#define REGISTER_NAMES \
{"r0","r1","r2","r3",  \
"r4","r5","r6","r7",   \
"r8","r9","r10","r11", \
"r12","r13","r14","r15"}

extern char *msp430_register_names[];

#define REGISTER_NAME(I) ((I>=0 || I<16 ) ? msp430_register_names[I] : NULL)


/* Extract from an array REGBUF containing the (raw) register state
    a function return value of type TYPE, and copy that, in virtual format,
    into VALBUF.  */ 
extern void msp430_extract_return_value PARAMS((struct type *, char *, char *));
#define EXTRACT_RETURN_VALUE(TYPE,REGBUF,VALBUF) \
     msp430_extract_return_value (TYPE, (char *)(REGBUF), (char *)(VALBUF))

/* the same, but Store... */
extern void msp430_store_return_value PARAMS((struct type *, char *));
#define STORE_RETURN_VALUE(TYPE,VALBUF) \
     msp430_store_return_value(TYPE, (char *) (VALBUF))
 

#define USE_STRUCT_CONVENTION(GCC_P, TYPE)  (sizeof(TYPE) > 8 ? 1 : 0)


/* When defined, extract from the array REGBUF (containing the raw
   register state) the ORE_ADDR' at which a function should return
   its structure value. */
#define STRUCT_RETURN_ADDRESS_USUALLY_IS_IN_R15	15
#define EXTRACT_STRUCT_VALUE_ADDRESS(REGBUF) \
extract_address(REGBUF + REGISTER_BYTE(STRUCT_RETURN_ADDRESS_USUALLY_IS_IN_R15), \
		REGISTER_RAW_SIZE(STRUCT_RETURN_ADDRESS_USUALLY_IS_IN_R15))


/* Given FRAME, return a pointer to the calling frame. */
CORE_ADDR msp430_frame_chain PARAMS ((struct frame_info *));
#define FRAME_CHAIN(FRAME) msp430_frame_chain(FRAME)


#define FRAME_CHAIN_VALID(fp,fi) func_frame_chain_valid (fp, fi)


/* A macro that tells us whether the function invocation represented
   by FI does not have a frame on the stack associated with it.  If it
   does not, FRAMELESS is set to 1, else 0.  */
#define FRAMELESS_FUNCTION_INVOCATION(FI) \
   (frameless_look_for_prologue(FI))

extern void msp430_init_extra_frame_info ();
#define INIT_EXTRA_FRAME_INFO(fromleaf, fi) \
	msp430_init_extra_frame_info (fromleaf, fi)


extern CORE_ADDR msp430_frame_saved_pc PARAMS ((struct frame_info *));

#define FRAME_SAVED_PC(FRAME) msp430_frame_saved_pc(FRAME)

extern CORE_ADDR frame_args_address PARAMS ((struct frame_info *));
#define FRAME_ARGS_ADDRESS(fi) frame_args_address(fi)

extern CORE_ADDR frame_locals_address PARAMS ((struct frame_info *));
#define FRAME_LOCALS_ADDRESS(fi) frame_locals_address(fi);


#define FRAME_FIND_SAVED_REGS(frame_info, frame_saved_regs)       \
   msp430_frame_find_saved_regs(frame_info, &(frame_saved_regs))



#define FRAME_NUM_ARGS(fi) (-1)
#define FRAME_ARGS_SKIP 0

extern void msp430_print_register_hook PARAMS ((int regno));
#define PRINT_REGISTER_HOOK(regno) msp430_print_register_hook(regno)


#define GDB_TARGET_IS_MSP430

#define NUM_REALREGS 	16
#define FP_REGNUM	4
#define AP_REGNUM	5
#define SP_REGNUM	1

/*
 * CALL_DUMMY stuff:
 */

extern CORE_ADDR msp430_call_dummy_address ();

#define USE_GENERIC_DUMMY_FRAMES 	1
#define CALL_DUMMY                    {}
#define CALL_DUMMY_LENGTH             (0)
#define CALL_DUMMY_ADDRESS()          msp430_call_dummy_address ()
#define CALL_DUMMY_LOCATION           AT_ENTRY_POINT
#define CALL_DUMMY_START_OFFSET       (0)   
#define CALL_DUMMY_BREAKPOINT_OFFSET  (0)

#define PC_IN_CALL_DUMMY(PC, SP, FP)  generic_pc_in_call_dummy (PC, SP, FP)
#define FIX_CALL_DUMMY(DUMMY, START_SP, FUNADDR, NARGS, ARGS, TYPE, GCCP)


#define PUSH_ARGUMENTS(NARGS, ARGS, SP, STRUCT_RETURN, STRUCT_ADDR) \
	(SP) = msp430_push_arguments (NARGS, ARGS, SP, STRUCT_RETURN, STRUCT_ADDR)


#define PUSH_DUMMY_FRAME      generic_push_dummy_frame ()
#define POP_FRAME             msp430_pop_frame ()

#define PUSH_RETURN_ADDRESS(PC, SP)   msp430_push_return_address (PC, SP)

#define GET_SAVED_REGISTER

/* Count of pushed registers. From r4 to r15 (inclusively) */
#define MAX_PUSHES 		12
/* Number of the last pushed register. r17 for current avr-gcc */
#define LAST_PUSHED_REGNO 	15


extern breakpoint_from_pc_fn msp430_breakpoint_from_pc;
#define BREAKPOINT_FROM_PC(PCPTR, LENPTR) msp430_breakpoint_from_pc (PCPTR, LENPTR)




struct sreg_bits_s
{
   char *name;
   unsigned short mask;
};

struct frame_extra_info
{
	CORE_ADDR regs[NUM_REGS];
	int num_saved_regs;
	CORE_ADDR from_pc;
	CORE_ADDR args_pointer;
	int locals_size;
	int framereg;
	int framesize;
};
































