/* Parameters for execution on a Motorola MCore.
   Copyright 1995, 1999, 2000 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA. */

#include "regcache.h"

/* The msp430 is little endian (by default) */
#define TARGET_BYTE_ORDER_DEFAULT LITTLE_ENDIAN

#define REGISTER_SIZE 2
#define MAX_REGISTER_RAW_SIZE 2

#define REGISTER_VIRTUAL_TYPE(REG) builtin_type_short

#define REGISTER_BYTE(REG) ((REG) * REGISTER_SIZE)
#define REGISTER_VIRTUAL_SIZE(REG) 2
#define REGISTER_RAW_SIZE(REG) 2

#define MAX_REGISTER_VIRTUAL_SIZE 8

#define REGISTER_BYTES (NUM_REGS * REGISTER_SIZE)

extern char *msp430_register_names[];
#define REGISTER_NAME(I) msp430_register_names[I]


#define NUM_REGS 16
#define PC_REGNUM 0
#define SP_REGNUM 1

extern int msp430_real_fp;

#define FP_REGNUM ((msp430_real_fp)?msp430_real_fp:1)

#define FIRST_ARGREG 15
#define LAST_ARGREG 12
#define RETVAL_REGNUM 15

#define TARGET_PTR_BIT		16
#define TARGET_CHAR_BIT		8
#define TARGET_INT_BIT		16
#define TARGET_DOUBLE_BIT	32



/* Offset from address of function to start of its code.
   Zero on most machines.  */
#define FUNCTION_START_OFFSET 0

#define DECR_PC_AFTER_BREAK 0

/* BREAKPOINT_FROM_PC uses the program counter value to determine
   the breakpoint that should be used. */
extern breakpoint_from_pc_fn msp430_breakpoint_from_pc;
#define BREAKPOINT_FROM_PC(PCPTR, LENPTR) msp430_breakpoint_from_pc (PCPTR, LENPTR)

#define INNER_THAN(LHS,RHS) ((LHS) < (RHS))

extern  CORE_ADDR msp430_saved_pc_after_call(struct frame_info *fi);
#define SAVED_PC_AFTER_CALL(FRAME)  msp430_saved_pc_after_call(FRAME)

struct frame_info;
struct type;
struct value;

extern void msp430_init_extra_frame_info (struct frame_info *fi);
#define INIT_EXTRA_FRAME_INFO(FROMLEAF, FI) msp430_init_extra_frame_info ((FI))
#define INIT_FRAME_PC		/* Not necessary */
#define FRAME_INIT_SAVED_REGS(FI)	/* handled by init_extra_frame_info */

extern CORE_ADDR msp430_frame_chain (struct frame_info *fi);
#define FRAME_CHAIN(FI) msp430_frame_chain ((FI))
#define FRAME_CHAIN_VALID(FP, FI) generic_file_frame_chain_valid ((FP), (FI))

extern CORE_ADDR msp430_frame_saved_pc (struct frame_info *);
#define FRAME_SAVED_PC(FI) (msp430_frame_saved_pc ((FI)))

/* Extracting/storing return values. */
extern void msp430_store_return_value (struct type *type, char *valbuf);
#define STORE_RETURN_VALUE(TYPE, VALBUF) msp430_store_return_value ((TYPE), (VALBUF))

extern void msp430_extract_return_value (struct type *type, char *regbut, char *valbuf);
#define EXTRACT_RETURN_VALUE(TYPE,REGBUF,VALBUF) \
    msp430_extract_return_value ((TYPE), (REGBUF), (VALBUF));

#define STORE_STRUCT_RETURN(ADDR, SP)	/* handled by msp430_push_arguments */

extern CORE_ADDR msp430_extract_struct_value_address (char *regbuf);
#define EXTRACT_STRUCT_VALUE_ADDRESS(REGBUF) \
    msp430_extract_struct_value_address (REGBUF)

extern CORE_ADDR msp430_skip_prologue (CORE_ADDR pc);
#define SKIP_PROLOGUE(PC) msp430_skip_prologue (PC)

#define FRAME_ARGS_SKIP 0
extern CORE_ADDR msp430_frame_args_address (struct frame_info *fi);
#define FRAME_ARGS_ADDRESS(FI) msp430_frame_args_address ((FI))
extern CORE_ADDR msp430_frame_locals_address (struct frame_info *fi);
#define FRAME_LOCALS_ADDRESS(FI) msp430_frame_locals_address ((FI))
#define FRAME_NUM_ARGS(FI) (-1)


extern void msp430_pop_frame (struct frame_info *fi);
#define POP_FRAME msp430_pop_frame (get_current_frame ())

#define USE_GENERIC_DUMMY_FRAMES 1
#define CALL_DUMMY                   {0}
#define CALL_DUMMY_START_OFFSET      (0)
#define CALL_DUMMY_BREAKPOINT_OFFSET (0)
#define CALL_DUMMY_LOCATION          AT_ENTRY_POINT
#define FIX_CALL_DUMMY(DUMMY, START, FUNADDR, NARGS, ARGS, TYPE, GCCP)
#define CALL_DUMMY_ADDRESS()         entry_point_address ()
#define SIZEOF_CALL_DUMMY_WORDS      0
#define SAVE_DUMMY_FRAME_TOS(SP)     generic_save_dummy_frame_tos (SP)

extern CORE_ADDR msp430_push_return_address (CORE_ADDR, CORE_ADDR);
#define PUSH_RETURN_ADDRESS(PC, SP)  msp430_push_return_address (PC, SP)

#define PUSH_DUMMY_FRAME	generic_push_dummy_frame ()

extern CORE_ADDR msp430_push_arguments (int, struct value **, CORE_ADDR,
				       unsigned char, CORE_ADDR);
#define PUSH_ARGUMENTS(NARGS, ARGS, SP, STRUCT_RETURN, STRUCT_ADDR) \
  (SP) = msp430_push_arguments (NARGS, ARGS, SP, STRUCT_RETURN, STRUCT_ADDR)

#define PC_IN_CALL_DUMMY(PC, SP, FP) generic_pc_in_call_dummy (PC, SP, FP)

#define REG_STRUCT_HAS_ADDR(GCC_P, TYPE) (sizeof(TYPE) > 8 ? 1 : 0)

extern use_struct_convention_fn msp430_use_struct_convention;
#define USE_STRUCT_CONVENTION(GCC_P, TYPE) msp430_use_struct_convention (GCC_P, TYPE)

/* override the default get_saved_register function with
   one that takes account of generic CALL_DUMMY frames */
#define GET_SAVED_REGISTER(raw_buffer, optimized, addrp, frame, regnum, lval) \
    generic_get_saved_register (raw_buffer, optimized, addrp, frame, regnum, lval)

/* Cons up virtual frame pointer for trace */
extern void msp430_virtual_frame_pointer (CORE_ADDR, long *, long *);
#define TARGET_VIRTUAL_FRAME_POINTER(PC, REGP, OFFP) \
	msp430_virtual_frame_pointer ((PC), (REGP), (OFFP))

/* For PE, gcc will tell us what th real type of
   arguments are when it promotes arguments. */
#define BELIEVE_PCC_PROMOTION 1
