/* Remote target glue for Avr monitor.

   Copyright 2002 Dmitry Diky diwil@mail.ru

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License  
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "gdbcore.h"
#include "target.h"
#include "monitor.h"
#include "serial.h"
#include "xmodem.h"

static char *msp430_regnames[NUM_REGS] = 
{ "r0",   "r1",  "r2",    "r3",   "r4",   "r5",   "r6",   "r7",
  "r8",   "r9",  "r10",   "r11",  "r12",  "r13",  "r14",  "r15"};
  

static void msp430_open PARAMS ((char *args, int from_tty));
int msp430_dis_asm_read_memory (bfd_vma memaddr, bfd_byte *myaddr,
                                int len, disassemble_info *info);
 
int (*orig_dis_asm_read_memory) (bfd_vma memaddr, bfd_byte *myaddr,
                                int len, disassemble_info *info);
 



static void
msp430_supply_register (regname, regnamelen, val, vallen)
char *regname;
int regnamelen;
char *val;
int vallen;
{
    int regno;  
    
    regno = -1; 
    if(*regname == 'r' || *regname == 'R')
    {
    	char *p;
    	int val;
    	regno = strtoul (regname+1, &p, 10);
    	if (p == regname+1)
    		error ("msp430_supply_register ():  bad value from monitor: %s.",
    						regname);
    	regname = p;
    }
    
    if (regno >= 0)
    	val = monitor_supply_register (regno, val);
}

static int hashmark = 1;      /* flag set by "set hash" */
extern struct monitor_ops msp430_cmds; /* fwd decl */


/* load elf file... no performance crap here. */
static void
msp430_load(desc, file, hashmark)
struct serial * desc;
char *file;  
int hashmark;
{
    bfd *abfd;  
    asection *s;
    CORE_ADDR load_offset;
    CORE_ADDR target_addr = 0;   
    unsigned long data_count = 0;
    int i;
    char buf[16];
    
    load_offset = 0;
    
    abfd = bfd_openr (file, 0);
    
    if (!abfd)
    {
    	printf_filtered ("Unable to open file %s\n", file);
    	return;
    }
    
    if (bfd_check_format (abfd, bfd_object) == 0)
    {
    	printf_filtered ("File is not an object file\n");
    	return;
    }
    
 
    for (s = abfd->sections; s; s = s->next)
    {
      if (s->flags & SEC_CODE)
      {
  	bfd_size_type section_size;
  	bfd_vma vma;
  
  	vma = bfd_get_section_vma (abfd, s) + load_offset;
  	section_size = bfd_section_size (abfd, s);
  
  	data_count += section_size;
  
  	printf_filtered ("%s\t: 0x%4x ... 0x%4x  ",
  			 bfd_get_section_name (abfd, s), vma,
  			 vma + section_size);
  	gdb_flush (gdb_stdout);
  
  	for (i = 0; i < section_size; i += sizeof (buf))
  	{
  	    int numbytes;
  	    char targ_buf[16];
  
  	    numbytes = min (sizeof (buf), section_size - i);
  
  	    read_memory (target_addr, targ_buf, numbytes);
  	    target_addr += numbytes;
  
  	    bfd_get_section_contents (abfd, s, buf, i, numbytes);
  
  	    if (memcmp (targ_buf, buf, numbytes))
  	    {
  		fprintf (stderr, "Difference in load file\n");
  	    }
  	    if (hashmark)
  	    {
  		putchar_unfiltered ('#');
  		gdb_flush (gdb_stdout);
  	    }
  	}			/* Per-packet (or S-record) loop */
  
 	putchar_unfiltered ('\n');
      }				/* Loadable sections */
    }
    
    return_to_top_level (RETURN_QUIT);
}


static void
msp430_open(args, from_tty)
char *args;
int from_tty;
{
   monitor_open (args, &msp430_cmds, from_tty);
}

int (*orig_xfer_memory) (); 

 
static int
msp430_xfer_memory (memaddr, myaddr, len, write, target)
CORE_ADDR memaddr;
char *myaddr;
int len;
int write;
struct target_ops *target;               /* ignored */
{
   /* We've got Newmannnnn ... */     
   return (orig_xfer_memory) (memaddr, myaddr, len, write, target);
}

  
int
msp430_dis_asm_read_memory (memaddr, myaddr, len, info)
bfd_vma memaddr;
bfd_byte *myaddr;
int len;
disassemble_info *info;
{
   return (*orig_dis_asm_read_memory) (memaddr, myaddr, len, info);
}
  



static struct target_ops msp430_ops;
static struct monitor_ops msp430_cmds ;

static char *msp430_inits[] = {"\n", NULL};



static void
init_msp430_cmds(void)
{
    msp430_cmds.flags = MO_CLR_BREAK_USES_ADDR; /* flags */
    msp430_cmds.init =			msp430_inits;
    msp430_cmds.cont =			"c\n";	/* continue command */
    msp430_cmds.step =   		"s\n";    /* single step */
    msp430_cmds.stop =   		"\003";   /* Interrupt char (^C) */
    msp430_cmds.set_break =   		"bp 0x%x\n"; /* set a breakpoint */     
    msp430_cmds.clr_break =   		"bc 0x%x\n"; /* clear a breakpoint */   
    msp430_cmds.clr_all_break = 	"bc -1\n"; /* clear all breakpoints */
    msp430_cmds.fill =   		NULL;     /* memory fill cmd */

    msp430_cmds.setmem.cmdb =       "wb 0x%x 0x%x\n"; /* setmem.cmdb (addr, value) */
    msp430_cmds.setmem.cmdw =       "ww 0x%x 0x%x\n";       /* setmem.cmdw (addr, value) */
    msp430_cmds.setmem.cmdl =       "wl 0x%x 0x%x\n";       /* setmem.cmdl (addr, value) */
    msp430_cmds.setmem.cmdll =      NULL; /* setmem.cmdll (addr, value) */
    msp430_cmds.setmem.resp_delim = "="; /* setreg.resp_delim */
    msp430_cmds.setmem.term =       NULL; /* setreg.term */
    msp430_cmds.setmem.term_cmd =   NULL; /* setreg.term_cmd */

    msp430_cmds.getmem.cmdb =       "rb 0x%x 0x%x\n"; /* getmem.cmdb (startaddr, endaddr) */
    msp430_cmds.getmem.cmdw =       "rw 0x%x 0x%x\n"; /* getmem.cmdw (startaddr, endaddr) */ 
    msp430_cmds.getmem.cmdl =       "rl 0x%x 0x%x\n"; /* getmem.cmdl (startaddr, endaddr) */ 
    msp430_cmds.getmem.cmdll =      NULL; /* getmem.cmdll (startaddr, endaddr) */
    msp430_cmds.getmem.resp_delim = "0x[0-9a-fA-F]+ 0x"; /* getmem.resp_delim */
    msp430_cmds.getmem.term =       NULL;       /* getmem.term */   
    msp430_cmds.getmem.term_cmd =   NULL; /* getmem.term_cmd */

    msp430_cmds.setreg.cmd =        "$%s=0x%x\n"; /* setreg.cmd (name, value) */
    msp430_cmds.setreg.resp_delim = "="; /* setreg.resp_delim */
    msp430_cmds.setreg.term =       NULL; /* setreg.term */
    msp430_cmds.setreg.term_cmd =   NULL; /* setreg.term_cmd */  

    msp430_cmds.getreg.cmd =        "$%s\n"; /* getreg.cmd (name) */
    msp430_cmds.getreg.resp_delim = "="; /* getreg.resp_delim */
    msp430_cmds.getreg.term =       NULL; /* getreg.term */
    msp430_cmds.getreg.term_cmd =   NULL; /* getreg.term_cmd */

    msp430_cmds.dump_registers =    "R\n"; /* dump_registers */

    msp430_cmds.register_pattern =  "\\(\\w+\\) ?=\\(0x[0-9a-fA-F]+\\b\\)";
    msp430_cmds.supply_register =   msp430_supply_register; /* supply_register */
    msp430_cmds.load_routine =      msp430_load; /* load routine */
    msp430_cmds.load =   	    NULL;     /* download command */  
    msp430_cmds.loadresp =  	    NULL;  /* load response (^Q) */
    msp430_cmds.prompt =    	    "(msp430) "; /* monitor command prompt */
    msp430_cmds.line_term = 	    ".";  /* end-of-line terminator */
    msp430_cmds.cmd_end =   	    NULL;  /* optional command terminator */
    msp430_cmds.target =    	    &msp430_ops; /* target operations */ 
    msp430_cmds.regnames =  	    msp430_regnames; /* register names */
    msp430_cmds.magic =   	    MONITOR_OPS_MAGIC        ;       /* magic */
}


void
_initialize_msp430 ()
{
   init_msp430_cmds() ;
   init_monitor_ops (&msp430_ops);
   
   msp430_ops.to_shortname = "msp430";
   msp430_ops.to_longname = "MSP430 simulator / JTAG interface";
   msp430_ops.to_doc = "Remotely connected msp430 microcontroller";
   msp430_ops.to_open = msp430_open;
   
   add_target (&msp430_ops);
   orig_xfer_memory = msp430_ops.to_xfer_memory;
   msp430_ops.to_xfer_memory = msp430_xfer_memory; 
   orig_dis_asm_read_memory = tm_print_insn_info.read_memory_func;
   
   tm_print_insn_info.read_memory_func = msp430_dis_asm_read_memory;
}
